
#include "defs.h"
#include "intl.h"
#include "padarea.h"

#include "recognize_kanji.h"

static GtkWidget* kp_toolbar_btn(gchar *id, GCallback cb, gpointer data, gchar *tooltip_text)
{
    GtkWidget *btn, *image;
    btn = gtk_button_new();

    image = gtk_image_new_from_stock(id, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_container_add(GTK_CONTAINER(btn), image);

    gtk_button_set_relief(GTK_BUTTON(btn), GTK_RELIEF_NONE);
    g_signal_connect(G_OBJECT(btn), "clicked", G_CALLBACK(cb), data);
    gtk_widget_set_tooltip_text(btn, tooltip_text);
    return btn;
}

static void kp_annotate_cb(GtkWidget *w, gpointer data)
{
    kp_padarea_annotate(KP_PADAREA(data), gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)));
}

static void kp_clear_cb(GtkWidget *w, gpointer data)
{
    KpPadArea *pa = KP_PADAREA(data);
    kp_padarea_clear(pa);
}

static void kp_undo_cb(GtkWidget *w, gpointer data)
{
    KpPadArea *pa = KP_PADAREA(data);
    kp_padarea_undo(pa);
}

static void kp_add_characters(GtkTreeView *list, gunichar2 *kanji)
{
    GtkListStore *store = GTK_LIST_STORE(gtk_tree_view_get_model(list));
    GtkTreeIter iter;
    gint i, h;
    GtkTreeViewColumn *column = gtk_tree_view_get_column(list, 0);
    gtk_list_store_clear(store);
    if(!kanji) return;
    gtk_tree_view_column_cell_get_size(column, NULL, NULL, NULL, NULL, &h);
    GtkAllocation allocation;
    gtk_widget_get_allocation(GTK_WIDGET(list), &allocation);
    for(i = 0; kanji[i]; i++)
    {
	if(h*(i + 2) > allocation.height)
	    break;
	gchar *utfstr = g_convert((gchar*)&(kanji[i]), 2, "UTF-8", "UTF-16LE", NULL, NULL, NULL);
	gtk_list_store_append(store, &iter);
	gtk_list_store_set(store, &iter, 0, utfstr, -1);
	g_free(utfstr);
    }
}

static void kp_selection_changed(GtkTreeSelection *selection, gpointer data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;
    gchar *sel;

    if(!gtk_tree_selection_get_selected(selection, &model, &iter))
	return;
    gtk_tree_model_get(model, &iter, 0, &sel, -1);
    gtk_clipboard_set_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY), sel, -1);
    gtk_clipboard_set_text(gtk_clipboard_get(NULL), sel, -1);
}

static void kp_stroke_add_remove_cb(KpPadArea *pa, gpointer data)
{
    GtkTreeView *ka = GTK_TREE_VIEW(data);
    GList *strokes = pa->strokes;
    gunichar2 *kanji = recognize_kanji(strokes);
    kp_add_characters(ka, kanji);
}

static gboolean accel_activate_cb(GtkAccelGroup *accelgroup, GObject *arg1, guint arg2, GdkModifierType arg3, gpointer user_data)
{
    gulong keyval = arg2;
    gulong mask = arg3;
    if((mask == 0x4) && (keyval == 0x71))
	gtk_main_quit();
    return FALSE;
}

int main (int argc, char *argv[])
{
    GtkWidget *wnd, *vbox, *hbox, *hbox1, *padarea, *karea, *annotate, *tmp;

    gtk_init (&argc, &argv);

    wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_resizable(GTK_WINDOW(wnd), TRUE);
    gtk_window_set_default_size(GTK_WINDOW(wnd), 400, 300);
    gtk_window_set_title(GTK_WINDOW(wnd), "kp123");
    g_signal_connect(G_OBJECT(wnd), "destroy", G_CALLBACK(gtk_main_quit), NULL);

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_container_add(GTK_CONTAINER(wnd), vbox);

    tmp = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(tmp), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(vbox), tmp, TRUE, TRUE, 0);

    hbox1 = gtk_hbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(tmp), hbox1);

    padarea = g_object_new(KP_TYPE_PADAREA, NULL);
    gtk_box_pack_start(GTK_BOX(hbox1), padarea, TRUE, TRUE, 0);

    tmp = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox1), tmp, FALSE, FALSE, 0);

    GtkListStore *store = gtk_list_store_new(1, G_TYPE_STRING);
    karea = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(karea), FALSE);
    gtk_tree_view_set_enable_search(GTK_TREE_VIEW(karea), FALSE);
    gtk_widget_set_size_request(karea, 50, -1);
    GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
    renderer->xalign = 0.5;
    renderer->yalign = 0.5;
    renderer->xpad = 5;
    renderer->ypad = 5;
    GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes("kanji", GTK_CELL_RENDERER(renderer), "text", 0, NULL);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(karea), column, -1);
    gtk_box_pack_start(GTK_BOX(hbox1), karea, FALSE, FALSE, 0);
    GtkTreeSelection *select = gtk_tree_view_get_selection(GTK_TREE_VIEW(karea));
    gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
    g_signal_connect(G_OBJECT(select), "changed", G_CALLBACK(kp_selection_changed), (gpointer)hbox1);
    g_signal_connect(G_OBJECT(padarea), "stroke_add_remove", G_CALLBACK(kp_stroke_add_remove_cb), (gpointer)karea);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 3);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    annotate = gtk_check_button_new_with_label(_("Annotate"));
    g_signal_connect(G_OBJECT(annotate), "toggled", G_CALLBACK(kp_annotate_cb), (gpointer)padarea);
    gtk_box_pack_start(GTK_BOX(hbox), annotate, FALSE, FALSE, 0);

    tmp = kp_toolbar_btn(GTK_STOCK_QUIT, G_CALLBACK(gtk_main_quit), NULL, _("Quit"));
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 20);

    tmp = kp_toolbar_btn(GTK_STOCK_CLEAR, G_CALLBACK(kp_clear_cb), (gpointer)padarea, _("Clear"));
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 20);

    tmp = kp_toolbar_btn(GTK_STOCK_UNDO, G_CALLBACK(kp_undo_cb), (gpointer)padarea, _("Undo"));
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 20);

    GtkAccelGroup *accel_group = gtk_accel_group_new();
    gtk_window_add_accel_group(GTK_WINDOW(wnd), accel_group);
    GClosure *closure = g_cclosure_new(G_CALLBACK(accel_activate_cb), NULL, NULL);
    gtk_accel_group_connect(accel_group, 0x71, 0x4, GTK_ACCEL_VISIBLE, closure);

    gtk_widget_show_all(wnd);

    gtk_main ();

    return 0;
}

