
#include "defs.h"

#include "recognize_stroke.h"
#include "recognize_kanji.h"
#include "recognize_extra.h"

#if 0
void print_stroke(GList *s)
{
    g_print("stroke len = %d\n", g_list_length(s));
    for(;s; s = g_list_next(s))
    {
	gint16 x = ((GdkPoint *)s->data)->x;
	gint16 y = ((GdkPoint *)s->data)->y;
	g_printf("%d %d\n", x, y);
    }
}

static void print_gunichar2(gunichar2 *entry, gint len)
{
    gint i;
    for(i = 0; i < len; i++)
    {
	gchar *tmp = (gchar*)entry;
	if(tmp[0] == 0xD)
	    break;
	gchar *utfstr = g_convert((gchar*)(entry++), 2, "UTF-8", "UTF-16LE", NULL, NULL, NULL);
	g_printf("%s", utfstr);
	g_free(utfstr);
    }
    g_printf("\n");
}
#endif

typedef struct _kanji_result
{
    gunichar2 uc;
    gint dist;
} kanji_result;

gint LevenshteinDistance(gint l1, gchar *s1, gint l2, gchar *s2)
{
    gchar *d = calloc((l1+1)*(l2+1)+1, sizeof(gchar));

    int i, j;
    for(i = 0; i <= l1; i++)
	d[i*(l2 + 1)] = i;
    for(j = 0; j <= l2; j++)
	d[j] = j;

    for(j = 1; j <= l2; j++)
    {
	for(i = 1; i <= l1; i++)
	{
	    if(s1[i-1] == s2[j-1])
		d[i*(l2 + 1) + j] = d[(i-1)*(l2 + 1) + j - 1];
	    else
	    {
		gint a = d[(i-1)*(l2 + 1) + j] + 1;
		gint b = d[i*(l2 + 1) + j - 1] + 1;
		gint c = d[(i-1)*(l2 + 1) + j - 1] + 1;
		d[i*(l2 + 1) + j] = (a > b) ? ((b > c) ? c : b) : ((a > c) ? c : a);
	    }
	}
    }
    gint ret = d[(l1+1)*(l2+1) - 1];
    g_free(d);
    return ret;
}

static kanji_result* rate_next_kanji(GList *strokes, gchar **sdata, gunichar2 *entry)
{
    kanji_result *res = calloc(1, sizeof(kanji_result));
    res->uc = *++entry;
    gunichar2 *bakptr = entry;
    gint i, j, l, l1 = 0, l2 = 0;
    for(i = 0; i < g_strv_length(sdata); i++)
	l1 += strlen(sdata[i]);
    gchar *s1 = calloc(l1, sizeof(gchar));
    for(i = 0, j = 0; i < g_strv_length(sdata); i++)
    {
	l = strlen(sdata[i]);
	g_memmove(&(s1[j]), sdata[i], l);
	j += l;
    }
    for(l2 = 0; g_ascii_isalpha((gchar)*++entry); l2++);
    gchar *s2 = calloc(l2, sizeof(gchar));
    entry = bakptr;
    for(i = 0; i < l2; s2[i++] = (gchar)(*++entry));
    res->dist += LevenshteinDistance(l1, s1, l2, s2);
    g_free(s1);
    g_free(s2);
    entry++;
    if(*entry == '|')
    {
	res->dist += pass_extra_filters(strokes, entry);
    }
    return res;
}

static gint kanji_results_compare(gpointer *ptr1, gpointer *ptr2)
{
    kanji_result *kr1 = (kanji_result*)*ptr1, *kr2 = (kanji_result*)*ptr2;
    if(kr1->dist > kr2->dist)
	return 1;
    if(kr1->dist < kr2->dist)
	return -1;
    return 0;
}

static gunichar2* find_next_entry(gchar *allkanji, gunichar2 *entry, gint allkanjilen, gunichar2 key)
{
    if(allkanji == (gchar*)entry)
    {
	++entry;
	if(*entry != key)
	    return find_next_entry(allkanji, entry, allkanjilen, key);
    }
    else
    {
	if((gchar*)entry - allkanji >= allkanjilen)
	    return 0;
	while((gchar*)entry - allkanji < allkanjilen)
	{
	    if(*++entry == '\n')
	    {
		entry++;
		if(*entry == key)
		    break;
		if(*entry > key)
		    return 0;
		++entry;
	    }
	}
    }
    return entry;
}

static gunichar2* pick_kanji(GList *strokes, gchar **sdata, gchar *allkanji, gint allkanjilen)
{
    const gint MAX_DISTANCE = 5;
    gint datalen = g_strv_length(sdata), i;
    gunichar2 key = 'A' + datalen - 1;
    gunichar2 *entry = (gunichar2*)allkanji;
    if(key > 'Z')
	return 0;

    entry = find_next_entry(allkanji, entry, allkanjilen, key);
    if(!entry)
	return 0;
    GPtrArray *arr = g_ptr_array_new();
    g_ptr_array_set_free_func(arr, g_free);
    for(;;)
    {
	kanji_result *res = rate_next_kanji(strokes, sdata, entry);
	g_ptr_array_add(arr, res);
	g_ptr_array_sort(arr, (GCompareFunc)kanji_results_compare);
	for(i = arr->len - 1; i >= 0; i--)
	{
	    kanji_result *res = g_ptr_array_index(arr, i);
	    if(res->dist > MAX_DISTANCE)
		g_ptr_array_remove_index(arr, i);
	    else
		break;
	}
	entry = find_next_entry(allkanji, entry, allkanjilen, key);
	if(!entry)
	    break;
    }
    gunichar2 *ret = calloc(arr->len + 1, sizeof(gunichar2));
    for(i = 0; i < arr->len; i++)
    {
	kanji_result *res = (kanji_result*)g_ptr_array_index(arr, i);
	ret[i] = res->uc;
    }
    g_ptr_array_free(arr, TRUE);
    return ret;
}

static GMappedFile* load_strokes_txt(gchar *dir)
{
    gchar *fname = g_build_filename(dir, "strokes.txt", NULL);
    GMappedFile *file = g_mapped_file_new(fname, FALSE, NULL);
    g_free(fname);
    return file;
}

gunichar2* recognize_kanji(GList *strokes)
{
    static gchar **sdata = NULL;
    static gint sdata_len = 0;
    static GList *tmp = NULL;
    gint strokes_len = g_list_length(strokes);
    if(!strokes_len)
    {
	int i;
	for(i = 0; i < sdata_len; g_free(sdata[i++]));
	g_free(sdata);
	sdata = NULL;
	sdata_len = 0;
	g_list_free(tmp);
	tmp = NULL;
	return 0;
    }
    if(strokes_len == sdata_len - 1)
    {
	g_free(sdata[sdata_len - 1]);
	tmp = g_list_remove(tmp, g_list_last(tmp)->data);
    }
    sdata = g_realloc(sdata, (strokes_len + 1)*sizeof(gchar*));
    if(strokes_len == sdata_len + 1)
    {
	GList *s = NULL;
	sdata[strokes_len - 1] = recognize_stroke(g_list_first(g_list_last(strokes)->data), &s);
	tmp = g_list_append(tmp, s);
    }
    sdata[strokes_len] = 0;
    sdata_len = strokes_len;
    GMappedFile *file = load_strokes_txt("");
    if(!file)
    {
#ifdef KP123_DATADIR
	file = load_strokes_txt(KP123_DATADIR);
#endif
    }
    if(!file)
    {
	g_printf("failed to load strokes.txt\n");
	return 0;
    }
    gint allkanjilen = g_mapped_file_get_length(file);
    gchar *allkanji = g_mapped_file_get_contents(file);
    gunichar2 *result = pick_kanji(tmp, sdata, allkanji, allkanjilen);
    g_mapped_file_unref(file);
    return result;
}

