
#include "defs.h"
#include <math.h>

#include "recognize_stroke.h"

float dist(GdkPoint *pt0, GdkPoint *pt1)
{
    return sqrt(pow(pt1->x - pt0->x, 2) + pow(pt1->y - pt0->y, 2));
}

static float angle(GdkPoint *pt0, GdkPoint *pt, GdkPoint *pt2)
{
    float d0 = dist(pt0, pt);
    float d2 = dist(pt, pt2);
    float d = dist(pt0, pt2);
    return acos((pow(d0, 2) + pow(d2, 2) - pow(d, 2))/(2*d0*d2));
}

gchar get_stroke_char(GdkPoint *p0, GdkPoint *p1)
{
    float x = p1->x - p0->x;
    float y = p1->y - p0->y;
    float a = acos(x/dist(p0, p1));
    if(a <= M_PI/8)
	return 'F';
    if(y < 0)
    {
	if((M_PI/8 < a) && (a < M_PI/2))
	    return 'I';
    }
    else
    {
	if((5*M_PI/8 < a) && (a < 7*M_PI/8))
	    return 'A';
	if((3*M_PI/8 <= a) && (a <= 5*M_PI/8))
	    return 'B';
	if((M_PI/8 < a) && (a < 3*M_PI/8))
	    return 'C';
    }
    return 0;
}

static float find_largest_angle(GList *s, GList **out)
{
    float a = 0, a1;
    GList *s0, *s1, *s2;
    *out = 0;
    if(g_list_length(s) < 3)
	return 0;
    for(s0 = s; s0 != g_list_last(s0); s0 = g_list_next(s0))
    {
	for(s1 = g_list_next(s0); s1 && (s1 != g_list_last(s1)); s1 = g_list_next(s1))
	{
	    for(s2 = g_list_next(s1); s2 && (s2 != g_list_last(s2)); s2 = g_list_next(s2))
	    {
		GdkPoint *pt0 = (GdkPoint*)s0->data;
		GdkPoint *pt = (GdkPoint*)s1->data;
		GdkPoint *pt2 = (GdkPoint*)s2->data;
		a1 = angle(pt0, pt, pt2);
		if(a1 > a)
		{
		    a = a1;
		    *out = s1;
		}
	    }
	}
    }
    return a;
}

static GList* filter_equal_points(GList *s)
{
    GdkPoint *p, *p2;
    for(;s != g_list_last(s); s = g_list_next(s))
    {
	p = (GdkPoint*)s->data;
	p2 = (GdkPoint*)g_list_next(s)->data;
	if((p->x == p2->x) && (p->y == p2->y))
	{
	    s = g_list_delete_link(s, s);
	}
    }
    return g_list_first(s);
}

/* let A, B, C be consecutive points
 * if angle ABC larger then predefined value
 * then remove B */
static GList* filter_points_by_angle(GList *s)
{
    const float MIN_MAX_ANGLE = 170*M_PI/180;
    while(1)
    {
	if(g_list_length(s) < 3)
	    return g_list_first(s);
	GList *pt = 0;
	if(find_largest_angle(s, &pt) > MIN_MAX_ANGLE)
	    s = g_list_delete_link(s, pt);
	else
	    break;
    }
    if(g_list_length(s) < 3)
	return g_list_first(s);
    GdkPoint *p0, *p, *p2;
    for(s = g_list_next(s); s != g_list_last(s); s = g_list_next(s))
    {
	if(g_list_length(s) < 3)
	    return g_list_first(s);
	p0 = (GdkPoint*)g_list_previous(s)->data;
	p = (GdkPoint*)s->data;
	p2 = (GdkPoint*)g_list_next(s)->data;
	gchar ch0 = get_stroke_char(p0, p);
	gchar ch1 = get_stroke_char(p, p2);
	if(ch0 == ch1)
	    s = g_list_delete_link(s, s);
    }
    return g_list_first(s);
}

static float find_average_and_smallest_dist(GList *s, GList **pt0, GList **pt1, float *pdmin)
{
    float d = 0, dt, dmin = -1;
    int i = 0;
    for(; s != g_list_last(s); s = g_list_next(s), i++)
    {
	GdkPoint *p0 = (GdkPoint*)s->data;
	GdkPoint *p1 = (GdkPoint*)g_list_next(s)->data;
	dt = dist(p0, p1);
	if((dmin < 0) || (dmin > dt))
	{
	    dmin = dt;
	    *pt0 = s;
	    *pt1 = g_list_next(s);
	}
	d += dt;
    }
    *pdmin = dmin;
    return d/i;
}

/* let A, B be consecutive points
 * if dist(A,B) less than a certain percentage of average dist
 * then remove B */
static GList* filter_points_by_dist(GList *s)
{
    const float MAX_MIN_DIST = 0.3;
    float d0, dmin;
    GList *pt0 = 0, *pt1 = 0;
    while(1)
    {
	if(g_list_length(s) < 3)
	    return g_list_first(s);
	d0 = MAX_MIN_DIST*find_average_and_smallest_dist(s, &pt0, &pt1, &dmin);
	if(dmin < d0)
	{
	    GList *del;
	    if(pt0 == s)
		del = pt1;
	    else if(pt1 == g_list_last(s))
		del = pt0;
	    else
	    {
		GdkPoint *p0 = (GdkPoint*)pt0->data;
		GdkPoint *p01 = (GdkPoint*)g_list_previous(pt0)->data;
		GdkPoint *p1 = (GdkPoint*)pt1->data;
		GdkPoint *p11 = (GdkPoint*)g_list_next(pt1)->data;
		float d0 = dist(p0, p01);
		float d1 = dist(p1, p11);
		if(d0 < d1)
		    del = pt0;
		else
		    del = pt1;
	    }
	    s = g_list_delete_link(s, del);
	}
	else
	    break;
    }
    return s;
}

gchar *get_stroke_str(GList *s)
{
    gchar *ret = calloc(g_list_length(s), sizeof(gchar));
    gchar prev = -1;
    gint len = 0;
    for(;s != g_list_last(s); s = g_list_next(s))
    {
	GdkPoint *p0 = (GdkPoint*)s->data;
	GdkPoint *p1 = (GdkPoint*)g_list_next(s)->data;
	gchar ch = get_stroke_char(p0, p1);
	if(!ch)
	    continue;
	if(ch != prev)
	{
	    if(len)
		ch = g_ascii_tolower(ch);
	    ret[len++] = ch;
	}
	prev = ch;
    }
    return ret;
}

gchar* recognize_stroke(GList *stroke, GList **out)
{
    *out = g_list_copy(stroke);
    *out = filter_equal_points(*out);
    *out = filter_points_by_angle(*out);
    *out = filter_points_by_dist(*out);
    return get_stroke_str(*out);
}

